"""
3D 几何工具
"""

from typing import Iterable, NamedTuple
import numpy as np
from scipy.spatial.transform import Rotation as R


class SElement(NamedTuple):
    """3D元素
    平移和旋转

    Args:
        translation: 平移向量
        rotation: 旋转四元数
    """

    translation: np.ndarray
    rotation: np.ndarray


def default_element() -> SElement:
    """初始化默认3D元素"""
    return SElement(
        translation=np.array([0.0, 0.0, 0.0]), rotation=np.array([0.0, 0.0, 0.0, 1.0])
    )


def homogeneous_to_element(homogeneous: np.ndarray) -> SElement:
    """将齐次矩阵转换为3D元素"""
    translation = homogeneous[:3, 3]
    rotation = R.from_matrix(homogeneous[:3, :3]).as_quat()
    return SElement(translation=translation, rotation=rotation)


def to_rotation_matrix(element: SElement) -> np.ndarray:
    """提取旋转矩阵"""
    return R.from_quat(element.rotation).as_matrix()


def to_homogeneous_matrix(element: SElement) -> np.ndarray:
    """将3D元素转换为齐次矩阵"""
    matrix = np.eye(4)
    matrix[:3, :3] = R.from_quat(element.rotation).as_matrix()
    matrix[:3, 3] = element.translation
    return matrix


class Range3D(NamedTuple):
    """3D范围"""

    x_min: float
    y_min: float
    z_min: float
    x_max: float
    y_max: float
    z_max: float


class Box3D(NamedTuple):
    element: SElement
    size: Range3D


def transform_pc(xyz: np.ndarray, matrix: np.ndarray) -> np.ndarray:
    """
    三维点 转换到box坐标系
    """
    xyz_homo = np.hstack([xyz, np.ones((xyz.shape[0], 1))])
    return (matrix @ xyz_homo.T).T[:, :3]


def in_box_mask(xyz: np.ndarray, box: Box3D) -> np.ndarray:
    """
    返回布尔掩码, 表点在box内。
    pc: (N, 3) ndarray
    box: Box3D
    """
    # 世界坐标系 -> box局部坐标系
    t_glo_box = to_homogeneous_matrix(box.element)
    t_box_glo = np.linalg.inv(t_glo_box)
    xyz_in_box_element = transform_pc(xyz=xyz, matrix=t_box_glo)

    x_idx = 0
    y_idx = 1
    z_idx = 2
    mask = np.ones((xyz_in_box_element.shape[0],), dtype=np.bool_)
    mask &= xyz_in_box_element[:, x_idx] >= box.size.x_min
    mask &= xyz_in_box_element[:, x_idx] < box.size.x_max
    mask &= xyz_in_box_element[:, y_idx] >= box.size.y_min
    mask &= xyz_in_box_element[:, y_idx] < box.size.y_max
    mask &= xyz_in_box_element[:, z_idx] >= box.size.z_min
    mask &= xyz_in_box_element[:, z_idx] < box.size.z_max
    return mask


def in_boxes_mask(xyz: np.ndarray, boxes: Iterable[Box3D]) -> np.ndarray:
    """
    返回布尔掩码，表示点是否落在任意一个 box 内。
    - 在任意 box 内 → True
    - 在所有 box 外 → False
    """

    mask = np.zeros((xyz.shape[0],), dtype=np.bool_)
    for box in boxes:
        mask |= in_box_mask(xyz=xyz, box=box)
    return mask



# /bin/env python
# -*- coding: utf-8 -*-
"""
@author: inklov3
@date: 2025-07-16
@description: molar function tools
"""

import base64
from bisect import bisect_left
import datetime
import hashlib
import json
import sys
import uuid
from pathlib import Path
from typing import Any, Callable, Generator, Iterable, Sequence, TextIO, Tuple
from urllib.parse import unquote, urlparse

import numpy as np
from tqdm import tqdm
from blake3 import blake3
from loguru import logger
from loguru._logger import Logger as LoguruLogger
from returns.result import safe


@safe
def mkdir(path: Path) -> None:
    """创建目录"""
    path.mkdir(parents=True, exist_ok=True)


@safe
def init_logger(level: str = "WARNING", logs_path: Path = Path("logs")) -> LoguruLogger:
    """
    初始化日志
    """
    print("Logger initializing...")
    if getattr(logger, "_configured", False):
        return logger
    format = "{time:YYYY-MM-DD HH:mm:ss}<level>|{level}|{message}</level>"

    logger.remove()
    logger.add(
        sink=sys.stdout,
        format=format,
        level=level,
    )
    if logs_path:
        logs_path.mkdir(parents=True, exist_ok=True)
        log_path = logs_path / f"{datetime.datetime.now():%Y-%m-%d_%H-%M-%S}.log"
        logger.add(
            str(log_path),
            format=format,
            level=level,
            enqueue=True,
            encoding="utf-8",
        )
    setattr(logger, "_configured", True)
    logger.info("Logger initialized")
    return logger


@safe
def log_message(logger: LoguruLogger, message: str, level: str = "INFO") -> None:
    logger.log(level, message)


@safe
def log_info(message: str) -> None:
    log_message(logger, message, "INFO")


@safe
def log_warning(message: str) -> None:
    log_message(logger, message, "WARNING")


@safe
def log_error(message: str) -> None:
    log_message(logger, message, "ERROR")


@safe
def log_debug(message: str) -> None:
    log_message(logger, message, "DEBUG")


def tqdm_loguru(iterable, level="INFO", **tqdm_kwargs):
    """
    使用 loguru 打印 tqdm 进度的函数式包装器

    Args:
        iterable: 可迭代对象
        level: loguru 日志级别 (默认 INFO)
        **tqdm_kwargs: 传给 tqdm 的其它参数
    """

    class TqdmToLoguru:
        def __init__(self, log_level):
            self.log_level = log_level

        def write(self, buf):
            buf = buf.strip()
            if buf:
                logger.log(self.log_level, buf)

        def flush(self):
            pass

    return tqdm(iterable, file=TqdmToLoguru(level), **tqdm_kwargs)


def read_json(file: Path) -> dict:
    """
    读json
    """
    return json.loads(file.read_text())


class NonSerializable:
    """标记字段为不可序列化：序列化时将跳过此字段"""

    def __repr__(self):
        return "<NonSerializable>"


NON_SERIALIZABLE = NonSerializable()


@safe
def serialize(obj):
    """
    序列化
    """
    if obj is None:
        return None
    if isinstance(obj, (str, int, float, bool)):
        return obj
    if callable(getattr(obj, "_asdict", None)):
        return serialize(obj._asdict())
    if isinstance(obj, (list, tuple, set)):
        return [serialize(v) for v in obj if v is not NON_SERIALIZABLE]
    if isinstance(obj, dict):
        return {k: serialize(v) for k, v in obj.items() if v is not NON_SERIALIZABLE}
    if hasattr(obj, "__dict__"):
        filtered = {
            k: v for k, v in vars(obj).items() if not isinstance(v, NonSerializable)
        }
        return serialize(filtered)
    if isinstance(obj, (np.generic,)):
        return obj.item()
    # NumPy 数组 -> Python 原生嵌套 list
    if isinstance(obj, np.ndarray):
        return obj.tolist()
    raise TypeError(f"Unsupported serialization type: {type(obj)}")


def write_json(data: any, file: Path, indent=2):
    """
    写json
    """
    with open(file, "w", encoding="utf8") as file:
        json.dump(data, file, ensure_ascii=False, indent=indent, default=serialize)


def to_json_string(data: any) -> str:
    return json.dumps(serialize(data), ensure_ascii=False)


def hash_string(s: str, method="md5") -> str:
    """
    字符串hash
    """
    h = hashlib.new(method)
    h.update(s.encode("utf-8"))
    return h.hexdigest()


def partial_hash(file: Path, chunk_size=4 * 1024 * 1024) -> str:
    """
    分块hash
    """
    size = file.stat().st_size
    md5 = hashlib.md5()

    ranges = []
    ranges.append((0, chunk_size))  # 开头

    if size > chunk_size * 2:
        middle = max((size // 2) - (chunk_size // 2), chunk_size)
        ranges.append((middle, chunk_size))  # 中间

    if size > chunk_size:
        ranges.append((size - chunk_size, chunk_size))  # 结尾

    with file.open("rb") as f:
        for offset, length in ranges:
            f.seek(offset)
            md5.update(f.read(length))

    return md5.hexdigest()


def get_uuid() -> str:
    """
    获取uuid
    """
    random_uuid = uuid.uuid4()
    return str(random_uuid)


def get_timestamp() -> str:
    """
    获取时间戳
    """
    time = datetime.datetime.now()
    return time.strftime("%Y.%m.%d-%H:%M:%S.%f")


def search_with_suffix(path: Path, suffixes: set[str]) -> Generator[Path, None, None]:
    """递归查找目录后缀文件"""
    for f in path.rglob(r"*"):
        if f.suffix not in suffixes:
            continue
        yield f


def url_to_path(url: str) -> Path:
    """url转换为路径"""
    url_parsed = urlparse(url)
    unquote_path = unquote(url_parsed.path)
    return Path(unquote_path)


def recursive_replace(s: str, replace_map: dict[str, str]) -> str:
    """假递归替换"""
    for k, v in replace_map.items():
        s = s.replace(k, v)
    return s


def recursive_clean(s: str, clean_words: Iterable[str]) -> str:
    """假递归清理"""
    return recursive_replace(s, {word: "" for word in clean_words})


def to_hash(data: bytes, scheme: str = "blake3") -> str:
    match scheme:
        case "blake3":
            return blake3(data).hexdigest()
        case "md5":
            return hashlib.md5(data).hexdigest()
        case _:
            raise NotImplementedError(f"Unsupported hash scheme: {scheme}")


def to_base64(data: bytes) -> str:
    """
    转换为base64
    """
    return base64.b64encode(data).decode("utf-8")


def sort_dict(data: dict, sort_rule: Callable[any, int]) -> dict:
    """
    排序字典
    """
    return dict(sorted(data.items(), key=lambda x: sort_rule(x)))


def find_nearest(sorted_seq: Sequence, target: Any) -> Tuple[Any, int]:
    """
    有序列中最近 target 元素

    :param sorted_seq: 已排序的序列 (必须支持索引操作)
    :param target: 待查询的值
    :return: (最接近的元素, 该元素的索引)
    """
    pos = bisect_left(sorted_seq, target)

    if pos == 0:
        return sorted_seq[0], 0
    if pos == len(sorted_seq):
        return sorted_seq[-1], len(sorted_seq) - 1

    before, after = sorted_seq[pos - 1], sorted_seq[pos]
    nearest = after if after - target < target - before else before
    idx = pos if nearest == after else pos - 1
    return nearest, idx



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# pylint: disable=invalid-name
"""
author: inklov3
date: 2025-07-16
description: 导出工具类
"""

from dataclasses import dataclass, field


@dataclass
class LabelMeLabel:
    """
    LabelMe标签
    """

    name: str
    shape: str
    group_id: int


@dataclass
class LabelMeFrame:
    """
    LabelMe帧
    """

    imagePath: str
    imageData: str | None = field(default=None)



"""
nuScenes导出适配器
"""

import base64
from enum import Enum
from itertools import groupby
from pathlib import Path
import shutil
from typing import Callable, NamedTuple
import uuid
from fyuneru.geometry3d import (
    Box3D,
    Range3D,
    SElement,
    to_homogeneous_matrix,
    transform_pc,
)
from fyuneru.lib import mkdir
import numpy as np


class NuScenesIndex(Enum):
    """nuScenes索引文件"""

    ATTRIBUTE = "attribute.json"
    # 标定token + 传感器 token信息
    CALIBRATED_SENSOR = "calibrated_sensor.json"
    CATEGORY = "category.json"
    EGO_POSE = "ego_pose.json"
    INSTANCE = "instance.json"
    LIDARSEG = "lidarseg.json"
    LOG = "log.json"
    MAP = "map.json"
    SAMPLE_ANNOTATION = "sample_annotation.json"
    # 多个sample data token + calibration token
    SAMPLE_DATA = "sample_data.json"
    # 时间戳+sample data token
    SAMPLE_TIME = "sample_time.json"
    SAMPLE = "sample.json"
    SCENE = "scene.json"
    # 传感器的token和channel名称
    SENSOR = "sensor.json"
    VISIBILITY = "visibility.json"
    SAMPLES_SUB = "samples"
    SWEEPS_SUB = "sweeps"
    LIDARSEG_SUB = "lidarseg"


class Modality(Enum):
    """传感器类型"""

    CAMERA = "camera"
    LIDAR = "lidar"
    RADAR = "radar"


class Sensor(NamedTuple):
    """传感器 sensor.json 实体"""

    token: str | None
    channel: str
    modality: str


class EgoPose(NamedTuple):
    """ego_pose.json 实体"""

    token: str | None
    timestamp: int
    translation: list[float]
    rotation: list[float]


class Log(NamedTuple):
    """log.json 实体"""

    token: str | None
    logfile: str
    vehicle: str
    data_captured: str
    location: str
    map_token: str


class Map(NamedTuple):
    """map.json 实体"""

    category: str
    token: str | None
    filename: str
    log_tokens: list[str]


class Sample(NamedTuple):
    """sample.json 实体"""

    token: str | None
    timestamp: int
    prev: str
    next: str
    scene_token: str
    data: dict[str, str]
    anns: list[str]


class Scene(NamedTuple):
    """scene.json 实体"""

    token: str
    log_token: str
    nbr_samples: int
    first_sample_token: str
    last_sample_token: str
    name: str
    description: str


class CalibratedSensor(NamedTuple):
    """calibrated_sensor.json 实体"""

    token: str
    sensor_token: str
    translation: list
    rotation: list
    camera_intrinsic: list


class Visibility(NamedTuple):
    """visibility.json 实体"""

    description: str
    token: str
    level: str


class SampleData(NamedTuple):
    """sample_data.json 实体"""

    token: str
    sample_token: str
    ego_pose_token: str
    calibrated_sensor_token: str
    filename: str
    fileformat: str
    is_key_frame: bool
    height: int
    width: int
    timestamp: int
    prev: str
    next: str
    sensor_modality: str
    channel: str


class Category(NamedTuple):
    """category.json 实体"""

    token: str
    name: str
    description: str
    index: int


class Lidarseg(NamedTuple):
    """lidarseg.json 实体"""

    token: str
    sample_data_token: str
    filename: str


class GtsAnnotation(NamedTuple):
    """sample_annotation.json 实体"""

    token: str
    timestamp: int
    gt_path: str
    next: str
    prev: str


class Attribute(NamedTuple):
    """attribute.json 实体"""

    token: str
    name: str
    description: str


class SampleAnnotation(NamedTuple):
    """sample_annotation.json 实体"""

    token: str
    sample_token: str
    category_name: str
    instance_token: str
    visibility_token: str
    attribute_tokens: list[str]
    translation: list
    size: list
    rotation: list
    prev: str
    next: str
    num_lidar_pts: int
    num_radar_pts: int


class Instance(NamedTuple):
    """instance.json 实体"""

    token: str
    category_token: str
    nbr_annotations: int
    first_annotation_token: str
    last_annotation_token: str


def to_sample(sensor_name: str, sensor_files: list[Path], nuscenes_root: Path):
    """
    转化sample目录
    """
    sample_dir = nuscenes_root / NuScenesIndex.SAMPLES_SUB.value / sensor_name
    mkdir(sample_dir)
    for sensor_file in sensor_files:
        shutil.copy(sensor_file, sample_dir / sensor_file.name)


def cal_sample_path(sensor_name: str, sensor_file: Path) -> Path:
    """计算sample路径"""
    return NuScenesIndex.SAMPLES_SUB.value / sensor_name / sensor_file.name


def get_default_category_index_map() -> dict[str, int]:
    """nuscenes官方label:id语义"""
    return {
        "noise": 0,
        "animal": 1,
        "human.pedestrian.adult": 2,
        "human.pedestrian.child": 3,
        "human.pedestrian.construction_worker": 4,
        "human.pedestrian.personal_mobility": 5,
        "human.pedestrian.police_officer": 6,
        "human.pedestrian.stroller": 7,
        "human.pedestrian.wheelchair": 8,
        "movable_object.barrier": 9,
        "movable_object.debris": 10,
        "movable_object.pushable_pullable": 11,
        "movable_object.trafficcone": 12,
        "static_object.bicycle_rack": 13,
        "vehicle.bicycle": 14,
        "vehicle.bus.bendy": 15,
        "vehicle.bus.rigid": 16,
        "vehicle.car": 17,
        "vehicle.construction": 18,
        "vehicle.emergency.ambulance": 19,
        "vehicle.emergency.police": 20,
        "vehicle.motorcycle": 21,
        "vehicle.trailer": 22,
        "vehicle.truck": 23,
        "flat.driveable_surface": 24,
        "flat.other": 25,
        "flat.sidewalk": 26,
        "flat.terrain": 27,
        "static.manmade": 28,
        "static.other": 29,
        "static.vegetation": 30,
        "vehicle.ego": 31,
    }


def get_default_category_s() -> list[Category]:
    """获取默认category列表"""
    default_category_index_map = get_default_category_index_map()
    return [
        Category(
            token=gen_token(seed=name),
            name=name,
            description="",
            index=index,
        )
        for name, index in default_category_index_map.items()
    ]


def gen_token(seed: str | None = None) -> str:
    """生成token"""
    if not seed:
        return uuid.uuid4().hex
    data = seed.encode("utf-8")
    return base64.b64encode(data).decode("utf-8")


class MolarIndex(NamedTuple):
    """molar索引"""

    task: str
    item: str
    sample: str
    channel: str
    label: str


def hash_molar_index(molar_index: MolarIndex) -> str:
    """hash平台索引"""
    task = molar_index.task
    item = molar_index.item
    sample = molar_index.sample if molar_index.sample else ""
    channel = molar_index.channel if molar_index.channel else ""
    label = molar_index.label if molar_index.label else ""
    return f"{task}_{item}_{sample}_{channel}_{label}"


def unhash_molar_index(molar_hash: str) -> MolarIndex:
    """unhash平台索引"""
    task, item, sample, channel, label = molar_hash.split("_")
    return MolarIndex(task=task, item=item, sample=sample, channel=channel, label=label)


def xyz_wxyz_to_homogeneous_matrix(xyz: np.ndarray, wxyz: np.ndarray) -> np.ndarray:
    """wxyz转homogeneous_matrix"""
    element = SElement(translation=xyz, rotation=wxyz2xyzw(wxyz))
    return to_homogeneous_matrix(element)


def point_cloud_sensor_to_ego(
    point_cloud_in_sensor: np.ndarray,
    calibrated_sensor: CalibratedSensor,
) -> np.ndarray:
    """点云传感器坐标系到ego坐标系"""
    t_ego_sensor = xyz_wxyz_to_homogeneous_matrix(
        xyz=calibrated_sensor.translation, wxyz=calibrated_sensor.rotation
    )
    xyz_in_sensor = point_cloud_in_sensor[:, :3]
    xyz_in_ego = transform_pc(xyz=xyz_in_sensor, matrix=t_ego_sensor)
    point_cloud_in_ego = np.hstack([xyz_in_ego, point_cloud_in_sensor[:, 3:]])
    return point_cloud_in_ego


def load_ego_point_cloud(
    sample_data: SampleData,
    calibrated_sensor: CalibratedSensor,
    nusc_root: Path,
    point_cloud_loader: Callable[[Path], np.ndarray],
) -> np.ndarray:
    """加载ego坐标系点云"""
    bin_file = nusc_root / sample_data.filename
    load_point_cloud = point_cloud_loader
    point_cloud_in_sensor = load_point_cloud(bin_file)
    point_cloud_in_ego = point_cloud_sensor_to_ego(
        point_cloud_in_sensor, calibrated_sensor
    )
    return point_cloud_in_ego


def xyzw2wxyz(quat_xyzw: np.ndarray) -> np.ndarray:
    """wxyz转xyzw"""
    return np.roll(quat_xyzw, 1)


def wxyz2xyzw(quat_wxyz: np.ndarray) -> np.ndarray:
    """xyzw转wxyz"""
    return np.roll(quat_wxyz, -1)


def xyz_wxyz_to_element(xyz: np.ndarray, wxyz: np.ndarray) -> SElement:
    """xyzw转element"""
    xyzw = wxyz2xyzw(wxyz)
    return SElement(translation=xyz, rotation=xyzw)


def point_cloud_to_ego_coord(
    point_cloud: np.ndarray, calibrated_sensor: CalibratedSensor
) -> np.ndarray:
    """定云传感器坐标转化自车系坐标"""
    xyz_in_sensor_coord = point_cloud[:, :3]
    t_ego_sensor = to_homogeneous_matrix(
        xyz_wxyz_to_element(
            xyz=calibrated_sensor.translation, wxyz=calibrated_sensor.rotation
        )
    )
    xyz_in_ego_sensor_coord = transform_pc(xyz=xyz_in_sensor_coord, matrix=t_ego_sensor)
    return np.hstack([xyz_in_ego_sensor_coord, point_cloud[:, 3:]])


def init_func_sample_data_point_cloud_loader(
    nusc_root: Path, point_cloud_loader: Callable[[Path], np.ndarray]
) -> Callable[[SampleData], np.ndarray]:
    """初始化sample_data点云加载器"""
    load_point_cloud = point_cloud_loader

    def sample_data_point_cloud_loader(sample_data: SampleData) -> np.ndarray:
        pc_file = nusc_root / sample_data.filename
        return load_point_cloud(pc_file)

    return sample_data_point_cloud_loader


def init_func_ego_point_cloud_loader(
    calibrated_sensor_s: list[CalibratedSensor],
    sample_data_point_cloud_loader: Callable[[SampleData], np.ndarray],
) -> Callable[[SampleData], np.ndarray]:
    """初始化ego点云加载器"""
    token_indexed_calibrated_sensor = {
        calibrated_sensor.token: calibrated_sensor
        for calibrated_sensor in calibrated_sensor_s
    }
    load_sample_data_point_cloud = sample_data_point_cloud_loader

    def ego_point_cloud_loader(sample_data: SampleData) -> np.ndarray:
        point_cloud_in_sensor = load_sample_data_point_cloud(sample_data)
        calibrated_sensor = token_indexed_calibrated_sensor[
            sample_data.calibrated_sensor_token
        ]
        point_cloud_in_ego = point_cloud_sensor_to_ego(
            point_cloud_in_sensor, calibrated_sensor
        )
        return point_cloud_in_ego

    return ego_point_cloud_loader


def init_func_instance_id_generator(
    log_instance_s: list[Instance],
) -> Callable[[Instance], int]:
    """初始化实例id生成器
    Args:
        log_instance_s: log单位内的instance
    Returns:
        instance_id_generator: 实例id生成器
    """
    instance_id_map = {
        instance.token: idx for idx, instance in enumerate(log_instance_s, 1)
    }

    def instance_id_generator(instance_token: str) -> int:
        return instance_id_map[instance_token]

    return instance_id_generator


def init_func_global_point_cloud_loader(
    ego_pose_s: list[EgoPose],
    ego_point_cloud_loader: Callable[[SampleData], np.ndarray],
) -> Callable[[SampleData], np.ndarray]:
    """初始化全局点云加载器"""
    token_indexed_ego_pose = {ego_pose.token: ego_pose for ego_pose in ego_pose_s}

    def global_point_cloud_loader(sample_data: SampleData) -> np.ndarray:
        point_cloud_in_ego = ego_point_cloud_loader(sample_data)
        ego_pose_token = sample_data.ego_pose_token
        ego_pose = token_indexed_ego_pose[ego_pose_token]
        t_map_ego = to_homogeneous_matrix(
            xyz_wxyz_to_element(xyz=ego_pose.translation, wxyz=ego_pose.rotation)
        )
        xyz_in_ego = point_cloud_in_ego[:, :3]
        xyz_in_map = transform_pc(xyz=xyz_in_ego, matrix=t_map_ego)
        point_cloud_in_map = np.hstack([xyz_in_map, point_cloud_in_ego[:, 3:]])
        return point_cloud_in_map

    return global_point_cloud_loader


def init_func_log_scene_getter(scene_s: list[Scene]) -> Callable[[str], list[Scene]]:
    """初始化log单位内的scene获取"""
    log_token_scene_grouper = groupby(iterable=scene_s, key=lambda x: x.log_token)
    log_token_scene_s_map = {
        log_token: list(values) for log_token, values in log_token_scene_grouper
    }

    def log_scene_getter(log_token: str) -> list[Scene]:
        return log_token_scene_s_map[log_token]

    return log_scene_getter


def init_func_scene_sample_getter(
    sample_s: list[Sample],
) -> Callable[[str], list[Sample]]:
    """初始化scene单位内的sample获取"""
    scene_sample_groupper = groupby(
        iterable=sample_s, key=lambda sample: sample.scene_token
    )
    scene_sample_s_map = {
        scene_token: list[Sample](values)
        for scene_token, values in scene_sample_groupper
    }

    def scene_sample_getter(scene_token: str) -> list[Sample]:
        return scene_sample_s_map[scene_token]

    return scene_sample_getter


def init_func_sample_data_getter(
    sample_data_s: list[SampleData],
) -> Callable[[str], SampleData]:
    """初始化sample_data获取"""
    sample_data_groupper = groupby(iterable=sample_data_s, key=lambda x: x.sample_token)
    sample_data_s_map = {
        sample_token: list[SampleData](values)
        for sample_token, values in sample_data_groupper
    }

    def sample_data_getter(sample_token: str) -> SampleData:
        return sample_data_s_map[sample_token]

    return sample_data_getter


def init_func_sample_annotation_getter(
    sample_annotation_s: list[SampleAnnotation],
) -> Callable[[str], list[SampleAnnotation]]:
    """初始化sample_annotation获取"""
    sample_annotation_groupper = groupby(
        iterable=sample_annotation_s, key=lambda x: x.sample_token
    )
    sample_annotation_s_map = {
        sample_token: list[SampleAnnotation](values)
        for sample_token, values in sample_annotation_groupper
    }

    def sample_annotation_getter(sample_token: str) -> list[SampleAnnotation]:
        return sample_annotation_s_map[sample_token]

    return sample_annotation_getter


class InstanceMask(NamedTuple):
    """实例掩码"""

    sample_data_token: str
    instance_index_token_map: dict[int, str]
    instance_mask: np.ndarray


def nusc_to_plat_size(nusc_size: np.ndarray) -> np.ndarray:
    """nuscenes w l h->plat l w h"""
    width_idx = 0
    length_idx = 1
    height_idx = 2
    return np.array(
        [nusc_size[length_idx], nusc_size[width_idx], nusc_size[height_idx]]
    )


def size_to_range(size: np.ndarray) -> Range3D:
    """size转range"""
    x_idx = 0
    y_idx = 1
    z_idx = 2
    x_min = -size[x_idx] / 2
    x_max = size[x_idx] / 2
    y_min = -size[y_idx] / 2
    y_max = size[y_idx] / 2
    z_min = -size[z_idx] / 2
    z_max = size[z_idx] / 2
    return Range3D(
        x_min=x_min, y_min=y_min, z_min=z_min, x_max=x_max, y_max=y_max, z_max=z_max
    )


def annotation_to_box(annotation: SampleAnnotation) -> Box3D:
    """nusc标签转 box3d"""
    xyz = annotation.translation
    wxyz = annotation.rotation
    element = xyz_wxyz_to_element(xyz, wxyz)
    wlh = annotation.size
    lwh = nusc_to_plat_size(wlh)
    range3d = size_to_range(lwh)
    return Box3D(element=element, size=range3d)


def init_func_no_target_point_loader(
    global_point_cloud_loader: Callable[[SampleData], np.ndarray],
    sample_data_instance_mask_getter: Callable[[str], InstanceMask],
) -> Callable[[SampleData], np.ndarray]:
    """初始化无目标点云加载器"""

    def no_target_point_loader(sample_data: SampleData) -> np.ndarray:
        global_point_cloud = global_point_cloud_loader(sample_data)
        instance_mask = sample_data_instance_mask_getter(sample_data.token)
        return global_point_cloud[~instance_mask.instance_mask]

    return no_target_point_loader


def init_func_lidarseg_getter(
    lidarseg_s: list[Lidarseg],
) -> Callable[[str], Lidarseg]:
    """初始化lidarseg获取"""
    lidarseg_groupper = groupby(iterable=lidarseg_s, key=lambda x: x.sample_data_token)
    lidarseg_s_map = {
        sample_data_token: list[Lidarseg](values)
        for sample_data_token, values in lidarseg_groupper
    }

    def lidarseg_getter(sample_data_token: str) -> Lidarseg:
        return lidarseg_s_map[sample_data_token]

    return lidarseg_getter


def init_func_lidarseg_bin_loader(
    segmentation_dtype: np.dtype,
    nusc_root: Path,
    nuscenes_version: str,
) -> Callable[[Lidarseg], np.ndarray]:
    """初始化lidarseg二进制加载器"""

    def lidarseg_bin_loader(lidarseg: Lidarseg) -> np.ndarray:
        lidarseg_file = (
            nusc_root
            / NuScenesIndex.LIDARSEG_SUB.value
            / nuscenes_version
            / lidarseg.filename
        )
        return np.fromfile(lidarseg_file, dtype=segmentation_dtype)

    return lidarseg_bin_loader



import base64
import gzip
from io import BytesIO
import json
from pathlib import Path
from typing import NamedTuple
from urllib.parse import unquote, urlparse

import numpy as np
from fyuneru.geometry3d import SElement
from returns.maybe import Maybe
from funcy import first


class TaskConfig(NamedTuple):
    uid: str
    domain_id: str
    name: str
    setting: dict


class Label(NamedTuple):
    uid: str
    id: Maybe[int]
    draw_type: str
    hash: str
    label: Maybe[str]
    frame_index: int
    lens_index: Maybe[int]
    points: Maybe[list]
    attributes: Maybe[dict]
    label_id_map: Maybe[str]
    points_in_box: Maybe[int]
    group: int | None


class Frame(NamedTuple):
    idx: int
    url: str
    imgUrls: Maybe[list[str]]
    location: Maybe[SElement]
    size: Maybe[dict]


class Item(NamedTuple):
    uid: str
    batch_uid: str
    labels: list[Label]
    frames: list[Frame]


def is_merge(urls: list[str]) -> bool:
    """
    urls 是否是合并任务
    """
    return len(urls) == 1


def extract_frames(item: dict) -> list[Frame]:
    info = item["info"]["info"]
    locations = info.get("locations", [])
    image_urls = (
        info.get("url")
        or info.get("imgUrls")
        or [location["imgUrls"] for location in locations]
    )
    urls = info.get("pcdUrls") or info.get("urls", None)
    # locations 出现可能是叠帧、重建
    if locations:
        if is_merge(urls):
            return [
                Frame(idx=idx, url=url, imgUrls=imgUrls, location=location, size=None)
                for idx, (url, imgUrls, location) in enumerate(
                    zip(urls * len(locations), image_urls, locations)
                )
            ]
        return [
            Frame(idx=idx, url=url, imgUrls=imgUrls, location=location, size=None)
            for idx, (url, imgUrls, location) in enumerate(
                zip(urls, image_urls, locations)
            )
        ]
    # 没有 locations 但是有 urls 单帧点云
    elif urls:
        return [
            Frame(idx=idx, url=url, imgUrls=imgUrls, location=None, size=None)
            for idx, (url, imgUrls) in enumerate(zip(urls, image_urls))
        ]
    # 没有 pcd 就是 2D 任务
    elif image_urls:
        size = info.get("size", None)
        return [
            Frame(idx=idx, url=url, imgUrls=None, location=None, size=size[idx])
            for idx, url in enumerate(image_urls)
        ]
    else:
        raise ValueError("Unknown task")


def extract_label(label: dict) -> Label:
    label_data = label["data"]
    uid = label["_id"]
    id = label_data.get("id", None)
    draw_type = label_data["drawType"]
    hash = label_data["hash"]
    label = label_data.get("label", None)
    frame_index = label_data["frameIndex"]
    lens_index = label_data.get("lensIndex", None)
    points = label_data.get("points", None)
    attributes = label_data.get("attributes", None)
    label_id_map = label_data.get("labelIdMap", None)
    points_in_box = label_data.get("pointsInBox", None)
    group = label_data.get("group", None)
    return Label(
        uid=uid,
        id=id,
        draw_type=draw_type,
        hash=hash,
        label=label,
        frame_index=frame_index,
        lens_index=lens_index,
        points=points,
        attributes=attributes,
        label_id_map=label_id_map,
        points_in_box=points_in_box,
        group=group,
    )


def extract_labels(item: dict) -> list[Label]:
    return [extract_label(label=label) for label in item["labels"]]


def parse_task_config(task: dict) -> TaskConfig:
    uid = task["_id"]
    domain_id = task["domainId"]
    name = task["name"]
    setting = task["setting"]
    return TaskConfig(uid=uid, domain_id=domain_id, name=name, setting=setting)


def parse_export_config(config: dict) -> dict:
    return config


def parse_item(item: dict) -> Item:
    uid = item["_id"]
    batch_uid = item["item"]["batchId"]
    labels = extract_labels(item)
    frames = extract_frames(item)

    return Item(uid=uid, batch_uid=batch_uid, labels=labels, frames=frames)


def parse_items(items: list[dict]) -> list[Item]:
    return [parse_item(item) for item in items]


class ExportTask(NamedTuple):
    task_config: TaskConfig
    export_config: dict
    items: list[Item]


def parse_origin(origin: dict) -> ExportTask:
    task = origin.get("task")
    config = origin.get("config")
    data = origin.get("data")

    export_config = parse_export_config(config)
    task_config: TaskConfig = parse_task_config(task)
    items: list[Item] = parse_items(data)

    return ExportTask(task_config=task_config, export_config=export_config, items=items)


def url_to_path(url: str) -> Path:
    parsed_url_path = urlparse(url).path
    unquote_path = unquote(parsed_url_path)
    return Path(unquote_path)


def calculate_resource_dst(sub_path: Path, dst_root: Path, level: int) -> Path:
    return dst_root / Path(*sub_path.parts[level:])


def build_frame_resource(frame: Frame, dst_root: Path, level: int) -> dict[Path, str]:
    path_url_dict = dict()
    main_resource_path = url_to_path(frame.url)
    path_url_dict[calculate_resource_dst(main_resource_path, dst_root, level)] = (
        frame.url
    )
    if frame.imgUrls:
        path_url_dict.update(
            {
                calculate_resource_dst(url_to_path(img_url), dst_root, level): img_url
                for img_url in frame.imgUrls
            }
        )
    return path_url_dict


def extract_xyz(label: Label) -> np.ndarray:
    """提取xyz"""
    points = label.points
    translation = slice(0, 3)
    return np.array(points[translation])


def extract_rpy(label: Label) -> np.ndarray:
    """提取rpy"""
    points = label.points
    rpy = slice(3, 6)
    return np.array(points[rpy])


def extract_size(label: Label) -> np.ndarray:
    """提取size"""
    points = label.points
    size = slice(6, 9)
    return np.array(points[size])


def extract_half_size(label: Label) -> np.ndarray:
    """提取half_size"""
    points = label.points
    # 兼容预设矩形
    if len(points) != 9:
        points = points.tolist() + [0] * (9 - len(points))
    box_lwh = extract_size(label)
    box_size = box_lwh / 2
    box_size = np.array(
        [
            -box_size[0],
            box_size[0],
            -box_size[1],
            box_size[1],
            -box_size[2],
            box_size[2],
        ]
    )
    return box_size


def decompress_and_decode(encoded_str):
    """解析labelIdMap"""

    # 1. Base64 解码
    compressed_data = base64.b64decode(encoded_str)

    # 2. Gzip 解压
    with gzip.GzipFile(fileobj=BytesIO(compressed_data), mode="rb") as f:
        decompressed_data = f.read().decode("utf-8")  # 解压后的字节流转为字符串

    # 3. JSON 解析
    scalar_ins_list = json.loads(decompressed_data)
    return scalar_ins_list


def get_id_semantic_map(
    frame_seg_labels: list[Label], alias_mean_map: dict[str, int]
) -> dict[int, int]:
    """获取id到semantic的映射"""
    id_semantic_map = {}
    for seg_label in frame_seg_labels:
        if seg_label.draw_type != "SEMANTIC_POINT":
            continue
        id_semantic_map[seg_label.id] = alias_mean_map[seg_label.label]
    return id_semantic_map


def extract_semantic_base(seg_labels: list[Label]) -> Label:
    """提取semantic_base"""
    return first(
        seg_label for seg_label in seg_labels if seg_label.draw_type == "SEMANTIC_BASE"
    )
